#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2021/12/26 11:25 下午
# @Author  : lijiarui
# @Email   : lijiarui15@jd.com
# @Site    : 
# @File    : nacos_client.py
# @Software: PyCharm

import atexit
import json
import re
import time
import random
import inspect
from copy import copy
from typing import Callable
from threading import Timer, RLock, Thread, Lock

import nacos
from py_nacos_client.logger import logger as _logger
from py_nacos_client.clienterror import  NoProvider
from py_nacos_client.utils import na_obj


"""
Status of instances
"""
INSTANCE_STATUS_UP: str = "UP"
INSTANCE_STATUS_DOWN: str = "DOWN"
INSTANCE_STATUS_STARTING: str = "STARTING"
INSTANCE_STATUS_OUT_OF_SERVICE: str = "OUT_OF_SERVICE"
INSTANCE_STATUS_UNKNOWN: str = "UNKNOWN"

"""
Action type of instances
"""
ACTION_TYPE_ADDED: str = "ADDED"
ACTION_TYPE_MODIFIED: str = "MODIFIED"
ACTION_TYPE_DELETED: str = "DELETED"

"""
The timeout seconds that all http request to the eureka server
"""
_DEFAULT_TIME_OUT = 5

"""
Default configurations
"""
_DEFAULT_ENCODING = "utf-8"


class HostsWrapper:

    def __init__(self, ip = "", port = "", weight = 0, healthy = False, **kwargs):
        """
        {
		"valid": true,
		"marked": false,
		"instanceId": "10.10.10.10-8888-DEFAULT-nacos.test.1",
		"port": 8888,
		"ip": "10.10.10.10",
		"weight": 1.0,
		"metadata": {}
	    }
        :param kwargs:
        """
        self.ip = ip
        self.port = port
        self.weight = weight
        self.healthy = healthy
        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def ip_str(self):
        return str(self.ip) + ":" + str(self.port)

    @property
    def is_valid(self):
        if self.healthy == "true" or self.healthy is True:
            return True
        else:
            return False

    def __eq__(self, other):
        return self.ip_str() == other.ip_str()



class ListenerBase:
    """
    subscribe callback deal
    modify client_ins content
    """
    def __init__(self, name, client_ins):
        self.listener_name = "default"
        self.client_ins = client_ins
        self.name = name


    def launch(self, *args):
        """
        订阅服务回调，修改目标服务注册信息
        :param args:
        :return:
        """
        local_sub_instance = args[1].instance
        status = args[0]
        new_host_obj = HostsWrapper(ip=local_sub_instance.pop('ip'), port=local_sub_instance.pop('port'),
                                    weight=local_sub_instance.pop('weight'), valid=local_sub_instance.pop('healthy'),
                                    **local_sub_instance
                                    )
        getattr(self, status.lower() + "_serverinfo")(new_host_obj)


    def added_serverinfo(self, new_host_obj):
        if new_host_obj.is_valid:
            self.client_ins.load_balance_dict[self.name]['ip_list'].append(new_host_obj)


    def modified_serverinfo(self, new_host_obj):
        if not new_host_obj.is_valid:
            self.deleted_serverinfo(new_host_obj)
        else:
            self.added_serverinfo(new_host_obj)

    def deleted_serverinfo(self, new_host_obj):
        ip_list = self.client_ins.load_balance_dict[self.name]['ip_list']
        if ip_list:
            for index, ip in enumerate(ip_list):
                if new_host_obj == ip:
                    ip_list.pop(index)
                    break



class JdPyNacosClient(nacos.NacosClient):
    """
    封装nacos客户端，减少参数的使用
    """

    _instance_lock = Lock()
    Register_already = set()

    def __new__(cls, *args, **kwargs):
        if not hasattr(JdPyNacosClient, "_instance"):
            with JdPyNacosClient._instance_lock:
                if not hasattr(JdPyNacosClient, "_instance"):
                    JdPyNacosClient._instance = object.__new__(cls)
        return JdPyNacosClient._instance

    def __init__(self, server_addresses, server_center=None):
        """
        :param server_addresses: nacos服务端单机
        :param server_list: nacos服务端集群
        """
        namespace = None
        if server_center:
            server_addresses = server_center.address
            namespace = server_center.namespace

        super().__init__(server_addresses, namespace=namespace,  endpoint=None,ak=None, sk=None, username=None, password=None)
        self.__heartbeat_interval = 3
        self.__heartbeat_timer = Timer(self.__heartbeat_interval, self.__auto_send_beat)
        self.__heartbeat_timer.daemon = True
        self.register_dict = {}
        self.instance_dict = {}
        self.load_balance_dict = {}
        _logger.info("[client-init] endpoint:%s, tenant:%s" % (self.current_server, self.namespace))

    def __register(self, instance) -> None:
        try:
            regs = self.add_naming_instance(instance.service_name, ip=instance.ip,
                                                   port=instance.port, group_name=instance.group_name)
            if regs:
                self.Register_already.add(instance)
                _logger.info("服务注册成功。端口号:%s" % instance.port)
        except BaseException as e:
            _logger.error("服务注册失败。端口号:%s" % instance.port, exc_info=True)


    def __deregister(self, instance = None) -> None:
        """
        批量注销
        :param instance:
        :return:
        """
        def do_deregister(instance):
            try:
                regs = self.remove_naming_instance(instance.service_name, ip=instance.ip,
                                                           port=instance.port, group_name=instance.group_name)
                if regs:
                    _logger.info("服务注销成功。")

            except:
                _logger.exception("服务注销失败", exc_info=True)
        map(do_deregister, self.Register_already) if not instance else do_deregister(instance)


    @staticmethod
    def __is_ip(ip_str):
        return re.match(r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$', ip_str)


    def __heart_beat(self) -> None:
        """
        send heart beat
        :return:
        """
        def do_beat(instance):

            self.register_dict["healthy"] = self.healthy = int(time.time())
            try:
                re = self.send_heartbeat(instance.service_name, instance.ip, port=instance.port)
                _logger.info(re)
                if (re['code'] != 10200):
                    self.register_dict["healthy"] = int(time.time()) - 10
                    self.__register(instance)

            except:
                self.register_dict["healthy"] = int(time.time()) - 10
                _logger.error("服务心跳维持失败！", exc_info=True)

        map(do_beat, self.Register_already) if len(self.Register_already) > 1 else do_beat(list(self.Register_already)[0])


    def __auto_send_beat(self) -> None:
        while True:
            self.__heart_beat()
            time.sleep(self.__heartbeat_interval)

    #TODO

    def __get_instance_addr(self, service_name):

        """
        调用nacos接口请求，返回可用地址,md5一致则不改变内存中的数据，否则变更新的实例信息
        仅获取
        :param serviceName:
        :param group:
        :param namespaceId:
        :return:
        """
        if service_name not in self.load_balance_dict:
            self.load_balance_dict[service_name] = {}
        try:
            re = self.list_naming_instance(service_name=service_name)
            msg = re['hosts']
            _logger.info(msg)
        except Exception:
            _logger.error("获取实例错误！", exc_info=True)
            msg = []
        hosts = []
        if not msg:
            raise Exception

        for item in msg:
            hosts.append({
                'ip': item['ip'],
                'port': item['port'],
                'healthy': item['healthy']
            })
        hosts_json = json.dumps(hosts)
        md5Content = self.get_md5(hosts_json)
        try:
            oldMd5 = self.load_balance_dict[service_name]["md5"]
        except KeyError:
            self.load_balance_dict[service_name]["md5"] = md5Content
            oldMd5 = ""
        if oldMd5 != md5Content:
            healthy_hosts = list()
            for host in msg:
                healthy_hosts.append(HostsWrapper(**host))
            self.load_balance_dict[service_name]["ip_list"] = healthy_hosts
            self.load_balance_dict[service_name]["index"] = 0
        return hosts


    def __get_and_subscribe_server(self, instance_or_name):
        """
        添加订阅
        :param service_obj:
        :return:
        """
        ret = self.__get_instance_addr(instance_or_name)
        self.subscribe([ListenerBase(instance_or_name, self), ], **{"service_name": instance_or_name})
        return ret


    def add_regist(self, instance):
        """
        手动调用
        :param instance:
        :return:
        """
        self.__register(instance)


    def remove_ins(self, instance):
        """
        手动调用
        :param instance:
        :return:
        """
        self.__deregister(instance)


    def start(self, instance = None) -> None:
        try:
            self.__register(instance)
            self.__heartbeat_timer.start()
        except:
            pass


    def cancel(self) -> None:
        if self.__heartbeat_timer.is_alive():
            self.__heartbeat_timer.cancel()
        self.__deregister()


    def load_balance_client_avr(self, server_name):
        """
        轮询，由index 返回一个IP
        :param instance:
        :return:
        """
        if server_name not in self.load_balance_dict:
            self.__get_and_subscribe_server(server_name)
        index = self.load_balance_dict[server_name]["index"]
        l = len(self.load_balance_dict[server_name]["ip_list"])
        if l == 0:
            _logger.error("无可用服务 serviceName: " + server_name)
            return ""
        if index >= l:
            self.load_balance_dict[server_name]["index"] = 1
            host_obj = self.load_balance_dict[server_name]["ip_list"][0]
        else:
            self.load_balance_dict[server_name]["index"]  = index + 1
            host_obj = self.load_balance_dict[server_name]["ip_list"][index]

        return host_obj.ip_str()


    def get_random_provider(self, server_name):
        """
        根据权重和是否健康获取一个provider
        :param kwargs:
        :return:
        """
        if server_name not in self.load_balance_dict:
            self.__get_and_subscribe_server(server_name)
        service_url_list = [host_obj for host_obj in self.load_balance_dict[server_name]["ip_list"] if host_obj.is_valid]
        if not service_url_list:
            raise NoProvider(data='can not find provider for %s'%server_name)

        total_weight = 0
        same_weight = True
        last_service_url = None
        for host_obj in service_url_list:
            total_weight += host_obj.weight
            if same_weight and last_service_url and last_service_url.weight !=  host_obj.weight:
                same_weight = False
            last_service_url = host_obj

        if total_weight > 0 and not same_weight:
            offset = random.randint(0, total_weight - 1)
            for host_obj in service_url_list:
                offset -= host_obj.weight
                if offset < 0:
                    return host_obj.ip_str()

        return random.choice(service_url_list).ip_str()




__cache_key = "default"
__cache_clients = {}
__cache_clients_lock = RLock()

def init(instance=None) -> JdPyNacosClient:
    """
    Initialize an Client object and put it to cache.
    Unlike using Client class that you need to start and stop the client object by yourself, this method
    will start the client automatically after the object created and stop it when the programe exist.
    read Client for more information for the parameters details.
    :return: client object
    """
    with __cache_clients_lock:
        if __cache_key in __cache_clients:
            _logger.warning("A client is already running, try to stop it and start the new one!")
            __cache_clients[__cache_key].stop()
            del __cache_clients[__cache_key]
        client = JdPyNacosClient('',server_center=na_obj)
        __cache_clients[__cache_key] = client
        client.start(instance)
        return client


def stop() -> None:
    client = get_client()
    if client is not None:
        client.cancel()

def get_client() -> JdPyNacosClient:
    with __cache_clients_lock:
        if __cache_key in __cache_clients:
            return __cache_clients[__cache_key]
        else:
            return None

@atexit.register
def _cleanup_before_exist():
    if len(__cache_clients) > 0:
        _logger.debug("cleaning up clients")
        for k, cli in __cache_clients.items():
            _logger.debug(f"try to stop cache client [{k}] this will also unregister this client from the eureka server")
            cli.cancel()


def client_warpper(method, server_name):
    def get_ip(f):
        def mainPro():
            cli = get_client()
            if not cli:
                return ''
            else:
                # if method is not Callable:
                #     return getattr(cli, method)(server_name)
                return method(server_name)

        mainPro.__name__ = f.__name__
        return mainPro

    return get_ip


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2021/12/28 3:42 下午
# @Author  : lijiarui
# @Email   : lijiarui15@jd.com
# @Site    : 
# @File    : adapter_server.py
# @Software: PyCharm
import os
import yaml
from py_nacos_client.utils import get_host_ip
from py_nacos_client.logger import logger
# 当前文件路径
pro_dir = os.path.split(os.path.realpath(__file__))[0]


father_path=os.path.abspath(os.path.dirname(pro_dir)+os.path.sep+".")
NacosLibYamlPath = os.path.realpath(os.path.join(pro_dir, "microservice.yaml"))

DEFAULT_GROUP = "DEFAULT_GROUP"
NAMESPACE = "PUBLIC"


class Adapter:
      def __init__(self, obj, adapted_methods):
          self.obj = obj
          self.__dict__.update(adapted_methods)
      def __str__(self):
          return str(self.obj)



class Instance:
    """
    服务实例
    """
    def __init__(self, name, ip = get_host_ip(), port = "8000", group_name = DEFAULT_GROUP, name_space = NAMESPACE, **kwargs):
        self.service_name = name
        self.ip = ip
        self.port = port
        self.group_name = group_name
        self.name_space = name_space
        self.port_list = []
        if kwargs.get("port_list"):
            self.port_list.extend(kwargs["port_list"])

    @property
    def has_more_port(self):
        return len(self.port_list) > 1

    def __call__(self, *args, **kwargs):
        return self.service_name


def load_active_content(path=''):
    """
    加载yaml文件内容
    :return:
    """
    with open(NacosLibYamlPath if not path else path) as f:
        server_config = yaml.load(f.read(), Loader=yaml.FullLoader)
    #kwargs can be use to customsize what you want
    SERVER_NAME = server_config.pop("SERVER_NAME")
    IP = server_config.pop("IP") if server_config.pop("IP") else get_host_ip()
    PORT = server_config.pop("PORT")
    if server_config:
        for k in server_config:
            server_config[k.lower()] = server_config.pop(k)
    new_instance = Instance(SERVER_NAME, IP, PORT, **server_config)
    return new_instance



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2021/12/26 8:25 下午
# @Author  : lijiarui
# @Email   : lijiarui15@jd.com
# @Site    : 
# @File    : clienterror.py
# @Software: PyCharm
"""
custom exception
"""

nacos_client_error = {}


class NacosClientError(RuntimeError):
    code = None
    message = None
    data = None

    def __init__(self, message=None, data=None, code=None):
        RuntimeError.__init__(self)
        self.message = message or self.message
        self.data = data
        self.code = code
        assert self.code, "Error without code is not allowed"

    def __str__(self):
        return "NacosClientError({code}) : {message}".format(
            code=self.code,
            message=str(self.message)
        )

    def __unicode__(self):
        return u"NacosClientError({code}) : {message}".format(
            code=self.code,
            message=str(self.message)
        )


class NoProvider(NacosClientError):
    code = 5050
    message = u'No provider name {0}'
    provide_name = u''

    def __init__(self, message=None, data=None):
        self.provide_name = data
        NacosClientError.__init__(self, message=self.message.format(data), data=data)


class ConnectionFail(NacosClientError):
    code = 504
    message = u'connect failed {0}'

    def __init__(self, message=None, data=None):
        message = self.message.format(data)
        NacosClientError.__init__(self, message=message, data=data)


class MethodNotFound(NacosClientError):
    code = -601
    message = u'The method does not exist / is not available'

    def __init__(self, message=None, data=None):
        NacosClientError.__init__(self, message=message, data=data)


nacos_client_error[NoProvider.code] = NoProvider
nacos_client_error[ConnectionFail.code] = ConnectionFail
nacos_client_error[MethodNotFound.code] = MethodNotFound



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2021/12/26 7:52 下午
# @Author  : lijiarui
# @Email   : lijiarui15@jd.com
# @Site    : 
# @File    : logger.py
# @Software: PyCharm
"""
cache logger for use client
optional output to log file or terminal
"""
import logging


def getLogger(name=None):
    logger = logging.getLogger(name)
    ch = logging.StreamHandler()
    formatter = logging.Formatter('[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s',
                                  datefmt='%Y-%m-%d %H:%M:%S')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.propagate = False

    return logger


logger = getLogger('nacos_client')


def init(level):
    logging.addLevelName(logging.CRITICAL + 10, 'OFF')
    logger.setLevel(logging.getLevelName(level))


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2022/1/3 7:42 下午
# @Author  : lijiarui
# @Email   : lijiarui15@jd.com
# @Site    : 
# @File    : utils.py
# @Software: PyCharm

import os
import sys
import configparser
import socket
from py_nacos_client.logger import logger
# 当前文件路径
pro_dir = os.path.split(os.path.realpath(__file__))[0]
# 在当前文件路径下查找.ini文件

configPath = os.path.join(pro_dir, "./nacos_config.ini")
if configPath not in sys.path:
    sys.path.append(configPath)

conf = configparser.ConfigParser()
conf.read(configPath)

class NacosCenter(object):
    """
    Nacos connect info
    """
    def __init__(self, address, namespace, **kwargs):
        self.address = address
        self.namespace = namespace
        if kwargs:
            for k, v in kwargs:
                setattr(self, k, v)



def back_nacos_server():
    try:
        default_conf = conf.sections()[0]
    except IndexError:
        return "INI content not exist"
    server_addr = conf.get(default_conf, "SERVER_ADDRESSES")
    namespace = conf.get(default_conf, "NAMESPACE")
    na_obj = NacosCenter(address=server_addr,namespace=namespace)
    return na_obj

na_obj = back_nacos_server()



_local_ip = None


def get_host_ip():
    """
    get current ip addr
    :return:
    """
    global _local_ip
    s = None
    try:
        if not _local_ip:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(('8.8.8.8', 80))
            _local_ip = s.getsockname()[0]
        return _local_ip
    except:
        logger.warn("ERROR when get local host ip by socket connect", exc_info=True)
    finally:
        if s:
            s.close()



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2022/1/10 4:08 下午
# @Author  : lijiarui
# @Email   : lijiarui15@jd.com
# @Site    : 
# @File    : setup.py
# @Software: PyCharm
import setuptools
import io
import os
import sys
from shutil import rmtree

class UploadCommand(setuptools.Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds...')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution...')
        os.system(
            '{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine...')
        os.system('twine upload dist/*')

        sys.exit()

here = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(here, 'README.md'), encoding='UTF-8') as f:
    long_description = '\n' + f.read()

setuptools.setup(
    name='py_nacos_client',
    version='0.1.4',
    author='lijiarui',
    packages=setuptools.find_packages(),
    include_package_data=True,
    description="Python client for Nacos.",
    python_requires='>=3',
    install_requires=[
        'nacos-sdk-python'
    ],
    cmdclass={
        'upload': UploadCommand,
    },
    keywords=['jd_nacos_py',],
    classifiers=[
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License'
    ],
)